/* Copyright (c) 2005-2006, Stefan Eilemann <eile@equalizergraphics.com> 
   All rights reserved. */

#ifndef EQBASE_THREAD_H
#define EQBASE_THREAD_H

#include <eq/base/base.h>
#include <eq/base/lock.h>

#include <vector>

namespace eqBase
{
    class Lock;
    class ExecutionListener;

    /**
     * An abstraction to create a new execution thread.
     * 
     * Depending on the type, a different implementation is used to create the
     * thread.
     */
    class Thread 
    {
    public:
        /** 
         * Constructs a new thread.
         * 
         * @param type the execution model to use for the thread.
         */
        Thread();

        /** Destructs the thread. */
        virtual ~Thread();

        /** 
         * Starts the thread.
         *
         * All thread listeners will be notified from within the thread, after
         * the thread was initialized successfully.
         * 
         * @return <code>true</code> if the thread was launched,
         *         <code>false</code> otherwise.
         * @sa init(), run()
         */
        bool start();

        /** 
         * The init function for the child thread.
         *
         * The parent thread will not be unlocked before this function has been
         * executed. If the thread initialisation fails, that is, this method
         * did return false, the thread will be stopped and the start() method
         * will return false.
         * 
         * @return the success value of the thread initialisation.
         */
        virtual bool init(){ return true; }

        /** 
         * The entry function for the child thread.
         * 
         * @return the return value of the child thread.
         */
        virtual void* run() = 0;

        /** 
         * Exits the child thread immediately.
         * 
         * This function does not return. It is only to be called from the child
         * thread. The thread listeners will be notified.
         *
         * @param retVal the return value of the thread.
         */
        virtual void exit( void* retVal = NULL );

        /** 
         * Cancels (stops) the child thread.
         *
         * This function is not to be called from the child thread.
         */
        void cancel();

        /** 
         * Waits for the exit of the child thread.
         *
         * @param retVal output value for the return value of the child, can be
         *               <code>NULL</code>.
         * @return <code>true</code> if the thread was joined,
         *         <code>false</code> otherwise.
         */
        bool join( void** retVal=NULL );

        /** 
         * Returns if the thread is stopped.
         * 
         * Note that the thread may be neither running nor stopped if it is
         * currently starting or stopping.
         *
         * @return <code>true</code> if the thread is stopped,
         * <code>false</code> if not.
         */
        bool isStopped() const { return ( _state == STATE_STOPPED ); }

        /** 
         * Returns if the thread is running.
         * 
         * Note that the thread may be neither running nor stopped if it is
         * currently starting or stopping.
         *
         * @return <code>true</code> if the thread is running,
         * <code>false</code> if not.
         */
        bool isRunning() const { return ( _state == STATE_RUNNING ); }

        /** 
         * Returns if this thread is the current (calling) thread.
         * 
         * @return <code>true</code> if the current thread has is the same
         *         thread as this thread, <code>false</code> if not.
         */
        bool isCurrent() const;

        /** 
         * Add a new thread state listener.
         * 
         * @param listener the listener.
         */
        static void addListener( ExecutionListener* listener );

    private:
        /** The current state of this thread. */
        enum State
        {
            STATE_STOPPED,
            STATE_STARTING, // start() in progress
            STATE_RUNNING,
            STATE_STOPPING  // child no longer active, join() not yet called
        };

        State _state;
        Lock  _syncChild;

        pthread_t _threadID;

        static void* runChild( void* arg );
        void        _runChild();

        // listener API
        static Lock                            _listenerLock;
        static std::vector<ExecutionListener*> _listeners;
        static pthread_key_t                   _cleanupKey;

        static pthread_key_t _createCleanupKey();
        void _installCleanupHandler();

        static void _notifyStarted();
        static void _notifyStopping( void* arg );
    };
}

#endif //EQBASE_THREAD_H
